/* Copyright (C) Stephen Chung, 1991-1992.  All rights reserved. */

/* Will take Jim Breen's kinfo listing (in ASCII format) and     */
/* create kinfo.1 and kinfo.2  Merge the two files together to   */
/* form kinfo.dat                                                */

#include <stdio.h>
#include <mem.h>

#define MAXLINE    256

typedef struct {
	unsigned char strokes;
	unsigned char bushu;
    unsigned short int nelson;
	unsigned char grade;
	unsigned char on:4;
	unsigned char kun:4;
	unsigned char meanings:4;
	unsigned char padding;
    long int offset;
} RECORD;

RECORD rec;

int nr_records = 0;
long int ReadingOffset = 0L;
long int StringOffset = 0L;
unsigned int nr_entries = 0;


ProcessReadings(fp, s)
FILE *fp;
unsigned char *s;
{
	int i, j, k;
	int nr_on = 0, nr_kun = 0, nr_meaning = 0;
	unsigned char reading[MAXLINE];
	unsigned char remainder[MAXLINE];

    rec.offset = ReadingOffset;

	for(;s[0];) {
		reading[0] = remainder[0] = '\0';

		if (s[0] == '{') {
            sscanf(s, "{%[^}]} %[^\n]", reading, remainder);
		} else {
			sscanf(s, "%s %[^\n]", reading, remainder);
		}
		strcpy(s, remainder);
		if (!reading[0]) continue;

		if (reading[0] == 0xa4) {/* Hiragana */
			nr_kun++;
			for (i = 0; reading[i]; i += 2) {
                if (reading[i] == '(' || reading[i] == ')') {    /* Okurikana */
                    putc('\x7f', fp);
                    i--;
                    ReadingOffset++;
                    continue;
                }

				if (reading[i] != 0xa4) {
                    printf("Line %d: Non-hiragana in KUN reading (%x%x)!\n",
                            nr_entries + 1, reading[i], reading[i+1]);
				}
				putc(reading[i+1] & 0x7f, fp);
				ReadingOffset++;
			}
            putc('\n', fp);
			ReadingOffset++;
		} else if (reading[0] == 0xa5) { /* Katakana */
			nr_on++;
			for (i = 0; reading[i]; i += 2) {
				if (reading[i] != 0xa5) {
                    printf("Line %d: Non-katakana in ON reading (%x%x)!\n",
                            nr_entries + 1, reading[i], reading[i+1]);
				}
				putc(reading[i+1] & 0x7f, fp);
				ReadingOffset++;
			}
			putc('\n', fp);
			ReadingOffset++;
		} else if (' ' <= reading[0] && reading[0] < 127) {
			nr_meaning++;

            j = strlen(reading);
            if (reading[j-1] == '}') reading[j-1] = '\0';
            if (reading[0] == '{') i = 1; else i = 0;

            for (; reading[i]; i++) {
                reading[i] &= 0x7f;
                if (reading[i] < ' ' || reading[i] >= 127)
                    printf("Line %d: Control character in meaning!\n", nr_entries + 1);
				putc(reading[i], fp);
                ReadingOffset++;
			}
            putc('\n', fp);
			ReadingOffset++;
		} else {
			printf("Line %d: Bad reading\n", nr_entries + 1);
		}
    }

    if (nr_on >= 16)
        printf("Line %d: Too many ON readings! (%d)\n", nr_entries + 1,
            nr_on);
    else rec.on = nr_on;

    if (nr_kun >= 16)
        printf("Line %d: Too many KUN readings! (%d)\n", nr_entries + 1,
            nr_kun);
    else rec.kun = nr_kun;

    if (nr_meaning >= 16)
        printf("Line %d: Too many meanings! (%d)\n", nr_entries + 1,
                nr_meaning);
    else rec.meanings = nr_meaning;

    if (!nr_on && !nr_kun && !nr_meaning) rec.offset = -1L;
}



main(argc, argv)
int argc;
char *argv[];
{
	int i, j, k;
	FILE *infp, *outfp1, *outfp2;
	char line[MAXLINE];
	char remainder[MAXLINE];
	char jiscode[50];
	char nelson[50];
	char bushu[50];
	char strokes[50];
	char grade[50];


    if (argc != 3) {
        fprintf(stderr, "Usage: mkkinfo  kanji-details-file  number-of-records\n");
		exit(1);
	}

	infp = fopen(argv[1], "rt");
	if (infp == NULL) {
		fprintf(stderr, "ERROR - Cannot open input file %s!\n", argv[1]);
		exit(1);
	}

	outfp1 = fopen("kinfo.1", "wb");
	if (outfp1 == NULL) {
		fprintf(stderr, "ERROR - Cannot open output file kinfo.1!\n");
		fclose(infp);
		exit(1);
	}

    outfp2 = fopen("kinfo.2", "wb");
	if (outfp2 == NULL) {
		fprintf(stderr, "ERROR - Cannot open output file kinfo.2!\n");
		fclose(infp);
		fclose(outfp1);
		exit(1);
	}


    nr_records = atoi(argv[2]);

    StringOffset = (long int) nr_records * (long int) sizeof(RECORD);

	while (fgets(line, MAXLINE, infp) != NULL) {
		jiscode[0] = nelson[0] = bushu[0] = strokes[0] = grade[0] =
		remainder[0] = '\0';

		sscanf(line, "%s %*s %s %s %s %s %[^\n]",
			jiscode, nelson, bushu, strokes, grade, remainder);
		
		if (bushu[0] != 'B') {
			printf("Line %d: bushu does not start with 'B'\n",
				nr_entries + 1);
		} else {
			i = atoi(bushu + 1);
            if (i < 0 || i > 255)
				printf("Line %d: bad bushu number (%d)\n",
					nr_entries + 1, i);

            rec.bushu = i;
		}

		if (strokes[0] != 'S') {
			printf("Line %d: strokes does not start with 'S'\n",
				nr_entries + 1);
		} else {
			i = atoi(strokes + 1);
            if (i < 0 || i > 255)
				printf("Line %d: bad strokes number (%d)\n",
					nr_entries + 1, i);

            rec.strokes = i;
		}

		if (nelson[0] != 'N') {
			printf("Line %d: Nelson does not start with 'N'\n",
				nr_entries + 1);
		} else {
			i = atoi(nelson + 1);
			if (i < 0)
				printf("Line %d: bad Nelson number (%d)\n",
					nr_entries + 1, i);

            rec.nelson = i;
		}

		if (grade[0] != 'G') {
			printf("Line %d: grade does not start with 'G'\n",
                nr_entries + 1);
		} else {
			i = atoi(grade + 1);
			if (i < 0 || i > 9) 
				printf("Line %d: bad grade number (%d)\n",
					nr_entries + 1, i);

            rec.grade = i;
		}

        rec.offset = 0L;

		ProcessReadings(outfp2, remainder);

        if (rec.offset < 0L) rec.offset = 0L;
        else rec.offset += StringOffset;

		fwrite(&rec, sizeof(RECORD), 1, outfp1);

		nr_entries++;
		fprintf(stderr, "%d\r", nr_entries);
	}

	fclose(infp);
	fclose(outfp2);
	fclose(outfp1);
}

